@@ -1,4 +1,4 @@
-# COPYRIGHT (C) 2021-2022 Nicotine+ Team
+# COPYRIGHT (C) 2021-2022 Nicotine+ Contributors
 #
 # GNU GENERAL PUBLIC LICENSE
 #    Version 3, 29 June 2007
@@ -17,6 +17,7 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 from pynicotine.pluginsystem import BasePlugin
+from pynicotine.slskmessages import UserStatus
 
 
 class Plugin(BasePlugin):
@@ -41,13 +42,13 @@ class Plugin(BasePlugin):
         if user not in self.settings['users']:
             return
 
-        if status <= 0:
+        if status == UserStatus.OFFLINE:
             self.user_statuses[user] = status
             return
 
-        previous_status = self.user_statuses.get(user, 0)
+        previous_status = self.user_statuses.get(user, UserStatus.OFFLINE)
 
-        if previous_status <= 0:
+        if previous_status == UserStatus.OFFLINE:
             # User was previously offline
             self.user_statuses[user] = status
             self.core.userbrowse.browse_user(user, switch_page=False)
