@@ -566,9 +566,8 @@ def write_file_and_backup(path, callback, protect=False):
 
     # Back up old file to path.old
     try:
-        if os.path.exists(path_encoded):
-            from shutil import copy2
-            copy2(path, path_old_encoded)
+        if os.path.exists(path_encoded) and os.stat(path_encoded).st_size > 0:
+            os.replace(path_encoded, path_old_encoded)
 
             if protect:
                 os.chmod(path_old_encoded, 0o600)
@@ -588,6 +587,10 @@ def write_file_and_backup(path, callback, protect=False):
         with open(path_encoded, "w", encoding="utf-8") as file_handle:
             callback(file_handle)
 
+            # Force write to file immediately in case of hard shutdown
+            file_handle.flush()
+            os.fsync(file_handle.fileno())
+
     except Exception as error:
         log.add(_("Unable to save file %(path)s: %(error)s"), {
             "path": path,
@@ -597,7 +600,7 @@ def write_file_and_backup(path, callback, protect=False):
         # Attempt to restore file
         try:
             if os.path.exists(path_old_encoded):
-                os.replace(path_old_encoded, path)
+                os.replace(path_old_encoded, path_encoded)
 
         except Exception as second_error:
             log.add(_("Unable to restore previous file %(path)s: %(error)s"), {
